/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#ifndef _CUDA_XFACE_H
#define _CUDA_XFACE_H

#if defined(HAVE_CUDA)

#include <util.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_GPU 4

typedef struct {
	char name[32];
	int32 compute_version_major;
	int32 compute_version_minor;
	int32 clock_speed; /* in kHz */
	int32 num_compute_units;
	int32 constant_mem_size;
	int32 shared_mem_size;
	uint32 global_mem_size;
	int32 registers_per_block;
	int32 max_threads_per_block;
	int32 can_overlap;
	int32 warp_size;
	int32 max_thread_dim[3];
	int32 max_grid_size[3];
	int32 has_timeout;
	CUdevice device_handle;
} gpu_info_t;

typedef struct {
	int32 num_gpu;
	gpu_info_t info[MAX_GPU];
} gpu_config_t;

char * cuGetErrorMessage(CUresult result);

void gpu_init(gpu_config_t *config);

#define CUDA_TRY(func) \
	{ 			 				\
		CUresult status = func;				\
		if (status != CUDA_SUCCESS) {			\
			printf("error (line %d): %s\n", __LINE__,\
				cuGetErrorMessage(status));	\
			exit(-1);				\
		}						\
	}

#define CUDA_ALIGN_PARAM(offset, alignment) \
	(offset) = ((offset) + (alignment) - 1) & ~((alignment) - 1)

#ifdef __cplusplus
}
#endif

#endif /* HAVE_CUDA */

#endif /* !_CUDA_XFACE_H_ */
