/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#ifndef _CUDARHO_H_
#define _CUDARHO_H_

#include <mp.h>
#include <cuda_xface.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void (*rng_callback_t)(uint32 which_rho, void *rng_state, mp_t *x);

typedef struct {
	mp_t target;

	mp_t mont_r;
	uint32 mont_w;

	uint32 num_functions;

	/* intermediate state */

	mp_t *curr_values;

	uint32 curr_iter;
	uint32 curr_limit;

	uint32 *main_array;

	uint32 *gcd_array;	/* MAX_MP_WORDS x num_functions */
	uint32 *accum_array;	/* MAX_MP_WORDS x num_functions */
	uint32 *inc_array;	/* MAX_MP_WORDS x num_functions */
	uint32 *x_array;	/* MAX_MP_WORDS x num_functions */
	uint32 *y_array;	/* MAX_MP_WORDS x num_functions */
	uint32 *mont_temp;	/* (MAX_MP_WORDS+1) x num_functions */

#ifdef HAVE_CUDA
	gpu_info_t gpu_info;
	CUcontext gpu_context;
	CUmodule gpu_module; 
	CUfunction gpu_kernel;
	CUfunction gpu_kernel2;
	CUdeviceptr gpu_const_data;
	CUdeviceptr gpu_iter_data;
	uint32 num_blocks;
#endif

} rho_t;

void rho_init(rho_t *r, char *input, 
		rng_callback_t prng_func, void *prng_data, 
		uint32 num_functions, uint32 which_gpu);

void rho_free(rho_t *r);

void rho_run(rho_t *r, uint32 num_iterations);

mp_t *rho_get_curr(rho_t *r);

#ifdef __cplusplus
}
#endif

#endif /* !_CUDARHO_H_ */
